#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
# $Revision: 1.8.5.30 $
# $Source: /project/arrayrep/src/htc/unix/HTCAgent.pm,v $
#

eval 'exec /opt/VRTSperl/bin/perl -I `pwd`/../../lib -S $0 ${1+"$@"}'
    if 0;

package HTCAgent;
use ArrayRep;
use ag_i18n_inc;
use File::Temp qw/ tempfile /;
use POSIX qw( uname WNOHANG );

@ISA = ("ArrayRep");

use strict;
use warnings;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);

@EXPORT = qw(
);
$VERSION = '1.00';

my ($pairdisplay, $pairsplit, $pairevtwait, $pairresync, $horctakeover);

my $vcs_home = $VCSHOME;

my $DEVNULL = "/dev/null";
my $HANOTIFY = "$VCSHOME/bin/hanotify";
my $current_time = 0;

if ( $^O =~ /^MSWin/i ){
    $DEVNULL = "nul";
    # Add path to hanotify in windows
    my $vcshome = $ag_i18n_inc::VCSHOME;
    $vcshome =~ s/\"//g;
    $HANOTIFY = "$vcshome\\bin\\hanotify";
}

%default = (
    resname => '',
    name    => '',
    home    => '',
    role    => '',
    status  => '',
    split   => '',
    user    => '',
    domain  => '',
    enc_pword => '',
    oldstatus => '',
    resource_type => '',
    isSSUSWritable => '',
);

sub new {
    delete $ENV{'HORCC_MRCF'};
    my($proto, $resname, $name, $home, $splittakeover, $user, $domain, $enc_pword) = @_;
    my $class = ref($proto) || $proto;
    my $self = { %default };
    $self->{'resname'} = $resname;
    $self->{'name'} = $name;
    $self->{'home'} = $home;
    $self->{'split'} = $splittakeover;
    $self->{'user'} = $user;
    $self->{'domain'} = $domain;
    $self->{'enc_pword'} = $enc_pword;
    $self->{'role'} = "";
    $self->{'fence'} = "";
    $self->{'start_time'} = $^T;
    $self->{'oldstatus'} = "";
    $self->{'resource_type'} = "HTC";
    $self->{'isSSUSWritable'} = 1;

    if ( $^O =~ /^MSWin/i )
    {
        $self->{'home'} = "\"" . Win32::GetShortPathName($home) . "\"";
        $pairdisplay = "$self->{'home'}\\pairdisplay";
        $pairsplit = "$self->{'home'}\\pairsplit";
        $pairevtwait = "$self->{'home'}\\pairevtwait";
        $pairresync = "$self->{'home'}\\pairresync";
        $horctakeover = "$self->{'home'}\\horctakeover";
    } else {
        $self->{'home'} = $home;
        $pairdisplay = "$self->{'home'}/pairdisplay";
        $pairsplit = "$self->{'home'}/pairsplit";
        $pairevtwait = "$self->{'home'}/pairevtwait";
        $pairresync = "$self->{'home'}/pairresync";
        $horctakeover = "$self->{'home'}/horctakeover";
    }

    $current_time = time;

    bless($self, $class);
    return $self;
}

sub DESTROY {}

#
# This feature is only available on unix versions
#
sub Rep_Tech_RPO {
    my ($ret, $home, $groupname, $instance, $percent_behind, $old_fence, $new_fence, $old_state, $new_state, $percent_reported, $bmp, $sidefile, $out, $computed);
	($ret, $home) = VCSAG_GET_ATTR_VALUE ("BaseDir", -1, 1, @ARGV);
	exit 0 if ($ret != $VCSAG_SUCCESS);
	($ret, $groupname) = VCSAG_GET_ATTR_VALUE ("GroupName", -1, 1, @ARGV);
	exit 0 if ($ret != $VCSAG_SUCCESS);
	($ret, $instance) = VCSAG_GET_ATTR_VALUE ("Instance", -1, 1, @ARGV);
	exit 0 if ($ret != $VCSAG_SUCCESS);
	my $max = 0;
    my $cmd = "$home"."/pairdisplay -g $groupname"." -I$instance"." -fcd -CLI -l";
    VCSAG_LOGDBG_MSG (3, "Rep_Tech_RPO : Issuing Command [$cmd]");
    my $cmd_output = `$cmd`;
    $computed = 0;
    if (!$?){
        my @out_split = split ('\n', $cmd_output);
        foreach my $line (@out_split){
            if ($line =~m /Fence/) {
                next;
            }
            my @line_split = split (' ', $line);
            $new_state = $line_split[7];
            $new_fence = $line_split[8];
            $percent_reported = $line_split[9];
            if (!$old_fence && $new_fence ne "-"){
                $old_fence = $new_fence;
                $old_state = $new_state;
            } elsif ($new_fence eq "-") {
                $out = "Rep_Tech_RPO: Unable to fetch the fence level cannot report anything from array.";
                VCSAG_LOG_MSG("I", "$out", 6801, "$out");
                return $out;
            }
            if ($old_fence){
				if ($new_fence eq $old_fence && $old_state eq $new_state){
					if ($new_state eq "COPY"){
						$percent_behind = 100 - $percent_reported;
						$computed = 1;
						if ($max < $percent_behind) {
							$max = $percent_behind;
						}
					} elsif ($new_state eq "PAIR"){
						if ($new_fence eq "DATA"){
							#BMP : The percent_reported is percentage completed operation. #
							$percent_behind = 100 - $percent_reported;
							$computed = 1;
							if ($max < $percent_behind){
								$max = $percent_behind;
							}
						} elsif ($new_fence eq "ASYNC"){
							# Side File:The percent_reported is the percentage left to transfer #
							# to S-VOL.
							$percent_behind = $percent_reported;
							$computed = 1;
							if ($max < $percent_behind){
								$max = $percent_behind;
							}
						} elsif ($new_fence eq "NEVER" ){
							$out = "Rep_Tech_RPO: Fence Level is set to NEVER. Unable to compute replication lag.";
							VCSAG_LOG_MSG("I", "$out", 6802, "$out");
							return $out;
						}
					}
				} else {
					$out = "Rep_Tech_RPO: Replication is in a mix of states cannot compute Replication lag.";
					VCSAG_LOG_MSG("I", "$out", 6803, "$out");
					return $out;
				}
            }
        }
    }
	if ($computed){
		if ($new_state eq "COPY"){
			$out = "Rep_Tech_RPO: Replication Lag is $max% as reported by HTC.";
		} elsif ($new_state eq "PAIR"){
			if ( $new_fence eq "DATA"){
				$out = "Rep_Tech_RPO: Percentage of BITMAP left as reported by HTC is $max%.";
			} elsif ($new_fence eq "ASYNC"){
				$out = "Rep_Tech_RPO: Maximum side file percentage used reported by HTC is $max%.";
			}
		}
		VCSAG_LOG_MSG("I", "$out", 6804, "$out");
		return $out;
	} else {
		$out = "Rep_Tech_RPO: Currently we compute Replication Lag only if all the devices are in COPY or PAIR state.";
		VCSAG_LOG_MSG("I", "$out", 6805, "$out");
		return $out;
	}
}

# pair_state Accepts a flag to indicate whether pairdisplay o/p should be logged.
sub pair_state {
    my ($self, $log_op, $rolemonitor) = @_;
    $self->{'role'} = "";
    $self->{'status'} = "";

    my $cmd = $pairdisplay;
    my $cmd_op = `$cmd -l -fc -CLI -g $self->{'name'}`;
    my @query = split('\n', $cmd_op);

    # Reading tmp_output_file in tmpquery array
    my @tmpquery;

    foreach my $q (@query) {
        if ($q =~ /Status/) {
            next;
        }
        my (@fields) = split('\s+', $q);
        if ($self->{'role'} eq "") {
            $self->{'role'} = $fields[8];
        } elsif ($self->{'role'} eq "P-VOL") {
            if ($fields[8] ne "P-VOL") {
                $self->{'role'} = "Invalid";
                last;
            }
        } elsif ($self->{'role'} eq "S-VOL") {
            if ($fields[8] ne "S-VOL") {
                $self->{'role'} = "Invalid";
                last;
            }
        } elsif ($self->{'role'} ne "SMPL") {
            $self->{'role'} = "Invalid";
        }
        if ($self->{'fence'} eq "") {
            $self->{'fence'} = $fields[10];
        } elsif ($self->{'fence'} ne $fields[10]) {
            $self->{'fence'} = "Invalid";
        }
        if ($self->{'status'} eq "") {
            $self->{'status'} = $fields[9];
        }
        if (($fields[9] eq "COPY") || ($fields[9] eq "PSUE") || ($fields[9] eq "SSUS")) {
            $self->{'status'} = $fields[9];
        }
        if ( ((defined $rolemonitor) && ($rolemonitor == 1)) && $self->{'role'} eq "S-VOL" && $self->{'status'} eq "SSUS" ) {
            unless(@tmpquery){
                # Creating a temporary file to the output stream of child process
                my $fh = File::Temp->new ();
                my $tmp_output_file = $fh->filename;
                close ($fh);

                # Saving STDOUT output stream handle in OLDOUT
                open (OLDOUT, ">&STDOUT") or VCSAG_LOGDBG_MSG(4, "Unable to save STDOUT handle in OLDOUT: $!");

                # Setting STDOUT to a temporary output file
                open (STDOUT, ">$tmp_output_file") or VCSAG_LOGDBG_MSG(4,"Can't open STDOUT for [$tmp_output_file] temporary file: $!","$tmp_output_file");

                # pairdisplay command with remote horcm
                my $tmpcmd_op = "$cmd" . " -fc -CLI -g " . $self->{'name'};

                # Creating  a child process that execute above pairdisplay command
                my $child_pid = fork();
                VCSAG_LOGDBG_MSG(4,"Error in creating child process that execute command [$tmpcmd_op]","$tmpcmd_op") unless defined $child_pid;

                if ($child_pid == 0){
                    exec($tmpcmd_op);
                    exit(0);
                }
                else {
                    my ($type) = $self->{'resource_type'};
                    if (0 == length $type) {
                        # Resource type was not set in the $self object
                        ($type) = split('\n',`$ag_i18n_inc::HARES -value $self->{'resname'} Type`);
                    }
                    # pairdisplay without "-l" command timeout is twice the value of
                    # remote RAID manager, timeout configured in horcm conifguration file
                    # MonitorTimeout recommended value should be more than twice the value
                    # of remote RAID manager timeout with some addition buffer time (~10sec)
                    my ($remote_horcm_timeout) = split('\n',`$ag_i18n_inc::HATYPE -value $type MonitorTimeout`);
                    $remote_horcm_timeout -= 10;
                    eval{
                        my $isChildStillAlive;
                        while ((($isChildStillAlive = waitpid($child_pid, WNOHANG)) != -1) && ($remote_horcm_timeout--)) {
                            sleep 1;
                        }
                        if ($isChildStillAlive != -1) {
                            kill HUP => $child_pid or VCSAG_LOGDBG_MSG(4, "Unable to kill process with pid [$child_pid]: $!","$child_pid");
                            waitpid($child_pid, 0);
                        }
                    };
                }

                # Closing STDOUT output stream
                close (STDOUT);

                # Resetting STDOUT stream to previous state
                open (STDOUT, ">&OLDOUT") or VCSAG_LOGDBG_MSG(4, "Unable to recover STDOUT handle to previous state: $!");

                # Closing OLDOUT handle
                close (OLDOUT);

                open (FILE, "<$tmp_output_file") or VCSAG_LOGDBG_MSG(4, "Unable to open [$tmp_output_file] file: $!","$tmp_output_file");
                push @tmpquery, $_ while (<FILE>);
                close (FILE);
            }

            if(@tmpquery){
                foreach my $tmpline (@tmpquery) {
                    my @tmpfields = split('\s+', $tmpline);
                    # Check the M flag of correspoding P-VOL volume
                    # if M flag is not equal to 'W' then current SSUS device is
                    # in readonly state
                    if ($tmpfields[0] eq $fields[0] && $tmpfields[1] eq $fields[1] && $tmpfields[8] eq "P-VOL" && $tmpfields[13] ne "W"){
                        VCSAG_LOG_MSG("W", "Found HTC volume group [$fields[0]] containing device [$fields[1]] in SSUS readonly state", 6806, "$fields[0]","$fields[1]");
                        $self->{'isSSUSWritable'} = 0;
                        last;
                    }
                }
            }
            else {
                $self->{'isSSUSWritable'} = -1;
            }

            # Exit for loop when we found any volume in SSUS readonly state
            if ($self->{'isSSUSWritable'} <= 0){
                last;
            }
        }
    }
    if (!$self->{'role'}) {
                $self->{'role'} = "Unknown";
    }
    if (($self->{'role'} eq "Invalid" || $self->{'role'} eq "Unknown" || $self->{'role'} eq "SMPL") && ((defined $log_op) && ($log_op == 1))) {
        VCSAG_LOG_MSG("N", "pairdisplay returned following states for devices of $self->{'name'}", 6807,  "$self->{'name'}");
        $cmd_op =~ s/%/P/g;
        VCSAG_LOG_MSG("N", "\n$cmd_op", 6808, "$cmd_op");
    }
}

sub get_state {
    my ($self, $log_op, $rolemonitor) = @_;
    $self->pair_state($log_op, $rolemonitor);
    VCSAG_LOG_MSG("N", "state of $self->{'name'} is $self->{'status'}; role is $self->{'role'}", 6809, "$self->{'name'}", "$self->{'status'}", "$self->{'role'}");
}


sub failover_required {
    my ($self) = shift;
    if ($self->{'role'} eq "Unknown") {
        return 10;
    }
    if ($self->{'role'} eq "Invalid") {
        return 2;
    }
    if ($self->{'role'} eq "SMPL") {
        return 6;
    }

    # S-VOL is writable in SSUS state; merely using HTC terminology here
    if (($self->{'role'} eq "S-VOL") && ($self->{'status'} eq "SSUS") && ($self->{'isSSUSWritable'} == 1)) {
        return 3;
    }

    if (($self->{'role'} eq "S-VOL") && ($self->{'status'} eq "SSUS") && ($self->{'isSSUSWritable'} == -1)) {
        return 13;
    }

    # S-VOL is read only in SSUS state; merely using HTC terminology here
    if (($self->{'role'} eq "S-VOL") && ($self->{'status'} eq "SSUS")) {
        return 12;
    }

    if (($self->{'role'} eq "S-VOL") && ($self->{'status'} eq "PAIR")) {
        VCSAG_LOG_MSG("N", "checking for replication link failure", 6810);

        my $cmd = $pairdisplay;
        my @query = split('\n', `$cmd -fc -CLI -g $self->{'name'} 2>&1`);
        foreach my $q (@query) {
            if ($q =~ /Status/) {
                next;
            }
            my (@fields) = split('\s+', $q);

            if ((defined $fields[8]) && ($fields[8] ne "P-VOL")) {
                next;
            }

            if ((defined $fields[9]) && ($fields[9] eq "PSUE")) {
                return 4;
            }

            if ((defined $fields[9]) && ($fields[9] eq "PAIR")) {
                return 14;
            }
        }
        #could not connect to remote horcm and remote devices are in unknown state
        return 5;
    }

    if (($self->{'role'} eq "S-VOL") && ($self->{'status'} eq "PSUE")) {
        return 11;
    }

    if (($self->{'role'} eq "S-VOL") && ($self->{'status'} ne "SSWS")) {
        return 1;
    }
	return 0;
}

sub wait_for_sync {
    my ($self) = shift;
    if (($self->{'role'} eq "S-VOL") && ($self->{'status'} eq "COPY")) {
        VCSAG_LOG_MSG("N", "COPY status detected; entry point will wait until all volumes are in PAIR status or entry point times out", 6840);
        my ($type) = $self->{'resource_type'};
        if (0 == length $type) {
            # Resource type was not set in the $self object
            ($type) = split('\n',`$ag_i18n_inc::HARES -value $self->{'resname'} Type`);
        }

        my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value $type OnlineTimeout`);
        my $toval;
        if ( $^O =~ /^MSWin/i ){
            $toval = $oto - 10;
        }
        else{
            $toval = $oto - 30;
        }
        my $cmd = $pairevtwait . " -g " . $self->{'name'} . " -s pair -t " . $toval;
        VCSAG_LOG_MSG("N", "issuing pairevtwait of device group $self->{'name'} with command: $cmd", 6812, "$self->{'name'}", "$cmd");
        my $ret = system("$cmd > $DEVNULL ");
        if ($ret) {
            VCSAG_LOG_MSG("E", "COPY did not complete before entry point timeout", 6813);
            exit(0);
        }
        #TBD: to SSWS from SSUS
        }
    return 0;
}

sub failover {
    my ($self) = @_;
    my ($type) = $self->{'resource_type'};
    if (0 == length $type) {
        # Resource type was not set in the $self object
        ($type) = split('\n',`$ag_i18n_inc::HARES -value $self->{'resname'} Type`);
    }
    # OnlineTimeout value of HTC type should be four times more than the value
    # of remote RAID manager timeout with some additional buffer time (~10sec)
    # as if device are in S-VOL PAIR state online operation runs
    # 1. pairdisplay without "-l" and
    # 2. horctakeover command
    my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value $type OnlineTimeout`);

    # calculating online elapsed time
    my $elapsed_time = time - $current_time;
    my $toval = $oto - $elapsed_time - 10;

    if ($toval < $elapsed_time) {
        VCSAG_LOG_MSG("E", "Insufficient value for OnlineTimeout to perform horctakeover. Please increase OnlineTimeout value and try again.", 6842);
        exit(0);
    }

    my $cmd = $horctakeover . " -t " . $toval . " -g " . $self->{'name'} . " 2>&1";
    VCSAG_LOG_MSG("N", "Issuing horctakeover of device group $self->{'name'} with command: $cmd", 6841, "$self->{'name'}", "$cmd");
    my $ret = system("$cmd");
    $ret = $ret / 256;
    VCSAG_LOG_MSG("N", "horctakeover was issued; return code is $ret", 6816, "$ret");

    if ($ret == 225 && $self->{'fence'} eq "NEVER" && $self->{'role'} eq "S-VOL") {
        # call pairresync and online the resource.
        my $cur = time();
        my $maxloop = ($cur - $self->{'start_time'}) / 5;
        my $numloop = 0;
        while ((($self->{'status'} eq "COPY") || ($self->{'status'} eq "PAIR")) && ($numloop < $maxloop)) {
            VCSAG_LOG_MSG("N", "waiting for COPY/PAIR to transition...", 6817);
            sleep(5);
            $numloop = $numloop + 1;
            $self->get_state(0,0);
        }
        if (($self->{'status'} eq "SSWS") || ($self->{'status'} eq "SSUS")) {
            my $cmdtxt = $pairresync . " -g " . $self->{'name'} . " -swaps";
            VCSAG_LOG_MSG("N", "issuing optional pairresync -swaps of device group $self->{'name'} in case array is up but RAID manager down with command: $cmdtxt", 6818, "$self->{'name'}", "$cmdtxt");
            system("$cmdtxt 2>&1");
            $self->create_lockfile();
            return 0;
        } else {
            # Probably the COPY/PAIR transition did not complete.
            VCSAG_LOG_MSG("E". "SVOL status in group $self->{'name'} did not change to PAIR. Exiting.", 6819, "$self->{'name'}");
            exit(0);
        }
    }

    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);

    #error code EX_EWSTOT (233)
    if ($ret == 233) {
        VCSAG_LOG_MSG("C", "horctakeover was issued; return code is $ret, this indicates that the horctakeover command timed out. The HTC agent has encountered a critical error while coming online. Please note that if replication is not in PAIR status, then it is essential that this issue be corrected prior to bringing $grp back online. Failure to do so may result in data integrity issues. Please check the CCI/engine logs for more details on recovering from the error.", 6820, "$ret", "$grp");
        $self->flush_freeze();
        return 0;
    }
    if ($self->{'split'} == 0) {
        if ($ret <= 1) {
            $self->create_lockfile();
            return 0;
        }
        $self->get_state(1,0);
        if ($self->{'role'} eq "P-VOL") {
            $self->create_lockfile();
            return 0;
        }
        # For all other return codes, flush and freeze the group.
        if ($ret <= 5) {
            VCSAG_LOG_MSG("W", "horctakeover cannot swap roles of primary and secondary in group $self->{'name'}. error code is $ret. Freezing the group.", 6821, "$self->{'name'}", "$ret");
        } else {
            VCSAG_LOG_MSG("W", "horctakeover failed for group $self->{'name'}. error code is $ret. Freezing the group.", 6822, "$self->{'name'}", "$ret");
        }
        $self->flush_freeze();
    } elsif ($self->{'split'} == 1) {
        if ($ret <= 5) {
            $self->create_lockfile();
            return 0;
        }
        $self->get_state(1,0);
        if ($self->{'role'} eq "P-VOL" || (($self->{'role'} eq "S-VOL") && ($self->{'status'} eq "SSWS"))) {
            $self->create_lockfile();
            return 0;
        }
        # For all other return codes, flush and freeze the group.
        VCSAG_LOG_MSG("W", "horctakeover failed; error code is $ret. Freezing the group.", 6823, "$ret");
        $self->flush_freeze();
    }
    return 1;
}

sub clearfd {
    my ($self) = @_;
    my (@lines) = split('\n', `$ag_i18n_inc::HARES -list TargetResName=$self->{'resname'} Type=HTCSnap`);
    if ($#lines == -1) {
        return 0;
    }

    my $ostype = (uname())[0];
    chomp $ostype;
    if ( $ostype =~ /AIX/i ){
        my (@depfields) = split('\n',`$ag_i18n_inc::HARES -dep $self->{'resname'}`);
        foreach my $depline (@depfields) {
            my (@deppair) = split( '\s+', $depline );
            if ( $deppair[2] eq $self->{'resname'} ) {
                my @tmp = split('\n',`$ag_i18n_inc::HARES -value $deppair[1] Type`);
                if ( $tmp[0] eq "LVMVG" ){
                    return 0;
                }
            }
        }
    }

    my (@nodefields) = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
    my ($fdres, $sys) = split('\s+', $lines[$#lines]);
    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $fdres Group`);
    my (@flines) = split('\n', `$ag_i18n_inc::HARES -list Type=HTCSnap Group=$grp`);
    my $prevres = "";
    foreach my $fline (@flines) {
        my ($fdres, $sys) = split('\s+', $fline);
        if (!$prevres || ($prevres ne $fdres)) {
            $prevres = $fdres;
            my $ret = system("$ag_i18n_inc::HARES -action $fdres clearvm -sys $nodefields[5]");
            if ($ret) {
                VCSAG_LOG_MSG("E", "cannot clear VM headers from fire drill", 6824);
                exit(0);
            }
        } else {
            next;
        }
    }
    return 0;
}

sub linkdown {
    my ($self) = @_;
    if (($self->{'role'} eq "P-VOL") && ($self->{'status'} eq "PSUE")) {
        return 1;
    }
    return 0;
}

sub linksuspended {
    my ($self) = @_;
    if (($self->{'role'} eq "P-VOL") && ($self->{'status'} eq "PSUS")) {
        return 1;
    }
    return 0;
}

sub linkrestored {
    my ($self) = @_;
    if (($self->{'role'} eq "P-VOL") && (($self->{'status'} ne "PSUE") && ($self->{'status'} ne "PSUS"))) {
        return 1;
    }
    return 0;
}

sub resynchronize {
    my ($self) = @_;
    my $cmd = $pairresync . " -g " . $self->{'name'};
    my $ret = system("$cmd > $DEVNULL 2>&1");
    if (!$ret) {
        VCSAG_LOG_MSG("N", "automatic link resynchronization of group $self->{'name'} succeeded with command with command: [$cmd]", 6825, "$self->{'name'}", "$cmd");
    }
    return 0;
}

sub freezetarget {
    my ($self, $user, $domain, $enc_pword) = @_;
    my ($val) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} TargetFrozen`);
    if ($val) {
        return 0;
    }
    my (@cluslist) = split('\n', `$ag_i18n_inc::HACLUS -list`);
    if ($#cluslist < 1) {
        return;
    }
    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    my (@lines) = split('\n', `$ag_i18n_inc::HAGRP -value $grp TFrozen -clus $cluslist[1]`);
    if ($#lines < 1) {
        return;
    }
    my (@fields) = split('\s+', $lines[1]);
    if ($fields[3] eq "1") {
        return;
    }

    my $ret = 1;
    if ( $^O =~ /^MSWin/i ){
        if(($user ne "") && ($domain ne "") && ($enc_pword ne ""))  {
            $ret = system("callcmd.exe \"hagrp -freeze $grp -clus $cluslist[1] \" $user $domain $enc_pword");
        }
        else {
            $ret = system("callcmd.exe \"hagrp -freeze $grp -clus $cluslist[1] \" ");
        }
    }
    else{
        $ret = system("$ag_i18n_inc::HAGRP -freeze $grp -clus $cluslist[1]");
    }
    if (!$ret) {
        if($self->{'status'} eq "PSUE") {
            VCSAG_LOG_MSG("N", "service group $grp in cluster $cluslist[1] was frozen because the replication link has failed; group will be unfrozen when data is resynchronized", 6826, "$grp", "$cluslist[1]");
        } else {
            VCSAG_LOG_MSG("N", "service group $grp in cluster $cluslist[1] was frozen because the replication link has been suspended; group will be unfrozen when data is resynchronized", 6827, "$grp", "$cluslist[1]");
        }
        my $val = $grp . ":" . $cluslist[1];
        system("$ag_i18n_inc::HARES -modify $self->{'resname'} TargetFrozen $val");
    }
    return 0;
}

sub flush_freeze {
    my ($self) = @_;

    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    # Get localsys nodename.
    my (@nodeid) = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
    my $sys = $nodeid[5];
    # Flush the group on localhost.
    my $ret = system("$ag_i18n_inc::HAGRP -flush $grp -sys $sys");

    # Freeze the group in the cluster.
    $ret = system("$ag_i18n_inc::HAGRP -freeze $grp");

    # If in GCO freeze on remote cluster also
    my (@cluslist) = split('\n', `$ag_i18n_inc::HACLUS -list`);
    if ($#cluslist < 1) {
        return;
    }

    if ( $^O =~ /^MSWin/i ){
        if(($self->{'user'} ne "") && ($self->{'domain'} ne "") && ($self->{'enc_pword'} ne ""))    {
            system("callcmd.exe \"hagrp -freeze $grp -clus $cluslist[1] \" $self->{'user'} $self->{'domain'}  $self->{'enc_pword'}");
        }
        else {
            system("callcmd.exe \"hagrp -freeze $grp -clus $cluslist[1] \" ");
        }
    }
    else{
        system("$ag_i18n_inc::HAGRP -freeze $grp -clus $cluslist[1]");
    }
    return 0;
}

sub unfreezetarget {
    my ($self, $user, $domain, $enc_pword) = @_;

    my ($val) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} TargetFrozen`);
    if (!$val) {
        return 0;
    }
    my ($grp, $clus) = split(':', $val);

    my (@lines) = split('\n', `$ag_i18n_inc::HAGRP -value $grp TFrozen -clus $clus`);
    if ($#lines < 1) {
        return 0;
    }

    my (@fields) = split('\s+', $lines[1]);
    if ($fields[3] eq "0") {
        system("$ag_i18n_inc::HARES -modify $self->{'resname'} TargetFrozen \"\"");
        return 0;
    }

    my $ret = 1;
    if ( $^O =~ /^MSWin/i ){
        if(($user ne "") && ($domain ne "") && ($enc_pword ne ""))  {
            $ret = system("callcmd.exe \"hagrp -unfreeze $grp -clus $clus \" $user $domain $enc_pword");
        }
        else {
            $ret = system("callcmd.exe \"hagrp -unfreeze $grp -clus $clus \" ");
        }
    }
    else{
        $ret = system("$ag_i18n_inc::HAGRP -unfreeze $grp -clus $clus");
    }

    if (!$ret) {
        VCSAG_LOG_MSG("N", "service group $grp in cluster $clus was unfrozen because either replication link has been resynchronized or FreezeSecondaryOnSplit is set to 0", 6836, "$grp", "$clus");
        system("$ag_i18n_inc::HARES -modify $self->{'resname'} TargetFrozen \"\"");
    }
    return 0;
}

#
# Return 1 if a parent resource of a parent service group is ONLINE
#

sub is_parent_online {
    my ($self) = shift;
    my ($p_res, $ret, $sys);
    my @p_res_list = ();
    my $is_online = 0;

    my (@nodeid) = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
    $self->{'nodename'} = $sys =  $nodeid[5];

    # Are there any resources that depend on this resource
    @p_res_list = $self->get_res_deps();

    if ($#p_res_list >= 0) { # Parent resources were found
        foreach $p_res (@p_res_list) {

            $ret = system("$ag_i18n_inc::HARES -wait $p_res Probed 1" . " -sys $sys -time 120 2>$DEVNULL");
            if ($ret == 0) {
                my ($st) = split('\n', `$ag_i18n_inc::HARES -state $p_res -sys $sys`);
                chomp $st;
                if ($st =~ /ONLINE/) {
                    $is_online = 1;
                    last;
                }
            }
        }
    } else { # No parent resources
        if ($self->is_parent_sg_res_online()) {
            $is_online = 1;
        }
    }
    return $is_online;
}

#
# Return 1 if at least one OnOff/OnOnly resource from a parent service
# group is online.
#

sub is_parent_sg_res_online {
    my ($self) = shift;
    my ($sg_name, $p_sg_name, $line, $ret);
    my @lines = ();
    my @fields = ();
    my @valid_dep_lines = ();
    my $is_online = 0;
    my $sys = $self->{'nodename'};

	#
	# Get parent service groups that have an online local/online global
	# dependency to our service group
	#

    $sg_name = `$ag_i18n_inc::HARES -value $self->{'resname'} Group`;
    chomp $sg_name;

    (@lines) = split('\n', `$ag_i18n_inc::HAGRP -dep $sg_name`);
    chomp @lines;

    # VCS specific code --->
    (@valid_dep_lines) = grep (/online[\s+](local|global)/, @lines );

    foreach $line (@valid_dep_lines) {
        (@fields) = split('\s+', $line);

        if ($fields[1] eq $sg_name) {
            $p_sg_name = $fields[0];
        } else {
            next;
        }

		#
		# If at least one OnOff or OnOnly probed resource in the ParentSG
		# has gone online, then the HTC resource can be online as well
		#

		$ret = system("$ag_i18n_inc::HAGRP -wait $p_sg_name Probed 1 -sys $sys" . " -time 120 2>$DEVNULL");

		if ($ret == 0) {
			my @p_on_res_list = $self->get_res_list($p_sg_name, "ONLINE");

			#
			# At least 1 online resource was found. Note: $#X == 0
			# implies @X has 1 element
			#
			if ($#p_on_res_list >= 0){

				#
				# We found at least one ONLINE resource. If it is of type
				# OnOff or OnOnly, we can continue to remain ONLINE
				#
				foreach my $p_res (@p_on_res_list) {
					my $p_type = `$ag_i18n_inc::HARES -value $p_res Type`;
					chomp $p_type;
					my $p_oprn_type = `$ag_i18n_inc::HATYPE -value $p_type Operations`;
					chomp $p_oprn_type;
					if ($p_oprn_type =~ /None/i ) {
						next;
					}

					$is_online = 1;
					last;
				}
			}
		}
		last if ($is_online == 1);
	}
	return $is_online;
}

#
# Returns the names of resources that belong to a given group and
# are in particular state
#
sub get_res_list {
    my ($self) = shift;
    my ($sg_name, $res_state) = @_;
    my @retarr = ();
    my @tmparr = ();
    my ($ret, $nodename, $resname);

    $ret = run_cmd("$ag_i18n_inc::HARES -list Group=$sg_name State=$res_state",
                \@tmparr, 1);

    if ($ret) {
        foreach (@tmparr) {
            /^#/ && next;
            ($resname, $nodename)  = split;

            if ((!grep(/^${resname}$/, @retarr)) &&
                ($nodename =~ /$self->{'nodename'}/)) {

                push (@retarr, $resname);
            }
        }
    }
    return @retarr;
}

#
# Returns the names of parent resources
#
sub get_res_deps {
    my ($self) = shift;

    my ($sg_name, $res_state) = @_;
    my @retarr = ();
    my @tmparr = ();
    my ($ret, $grpname, $p_resname, $c_resname);

    $ret = run_cmd("$ag_i18n_inc::HARES -dep $self->{'resname'}", \@tmparr, 1);

    if ($ret) {
        foreach (@tmparr) {
            /^#/ && next;
            ($grpname, $p_resname, $c_resname)  = split;

			#
			# Note: with VCSOne, the 'resname' would be <grpname>.<resname>
			# whereas the $c_resname would only have the trailing <resname>
			# part
			#

			if ((!grep(/^${p_resname}$/, @retarr)) && ( $c_resname =~ /$self->{'resname'}$/ )) {
				push (@retarr, $p_resname);
			}
		}
	}
	return @retarr;
}

sub run_cmd {
    my ($cmd, $output_ref, $ignore_failure) = @_;
    my $retval = 1;

    @{$output_ref} = `$cmd`;

    if ($? != 0) {
        $retval = 0;
        unless (1 == $ignore_failure ) {
            VCSAG_LOG_MSG("W", "Command: \'$cmd\' failed !", 6829, "$cmd");
        }
    } else {
        chomp @{$output_ref};
    }

    return $retval;
}

sub generateTrap {
    my ($self) = @_;

    if($self->{'oldstatus'} eq $self->{'status'}) {
        return 0;
    }

    my ($devgrp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} GroupName`);
    my ($type) = $self->{'resource_type'};
    if (0 == length $type) {
        # Resource type was not set in the $self object
        ($type) = split('\n',`$ag_i18n_inc::HARES -value $self->{'resname'} Type`);
    }

    if($self->{'status'} eq "PSUE") {
        my $ret = system("$HANOTIFY -i 1.3.6.1.4.1.1302.3.8.10.2.1.8 -l Error -n $self->{'resname'} -T 0 -t \"$type Agent\" -m \"The state of $self->{'role'} devices in device group $devgrp has changed from $self->{'oldstatus'} to $self->{'status'}\"");
        VCSAG_LOG_MSG("C", "The state of $self->{'role'} devices in device group $devgrp has changed from $self->{'oldstatus'} to $self->{'status'}.", 6830, "$self->{'role'}", "$devgrp", "$self->{'oldstatus'}", "$self->{'status'}");
    }
    else {
        my $ret = system("$HANOTIFY -i 1.3.6.1.4.1.1302.3.8.10.2.1.9 -l Information -n $self->{'resname'} -T 0 -t \"$type Agent\"  -m \"The state of $self->{'role'} devices in device group $devgrp has changed from $self->{'oldstatus'} to $self->{'status'}\"");
        VCSAG_LOG_MSG("I", "The state of $self->{'role'} devices in device group $devgrp has changed from $self->{'oldstatus'} to $self->{'status'}", 6831, "$self->{'role'}", "$devgrp", "$self->{'oldstatus'}", "$self->{'status'}");
    }
}

sub initOldState {
	my ($self) = @_;
    my ($type) = $self->{'resource_type'};
    if (0 == length $type) {
        # Resource type was not set in the $self object
        ($type) = split('\n',`$ag_i18n_inc::HARES -value $self->{'resname'} Type`);
    }

    my $oldstatefile;
    if ( $^O =~ /^MSWin/i ){
        my $vcshome = $ag_i18n_inc::VCSHOME;
        $vcshome =~ s/\"//g;
        $oldstatefile = "$vcshome" . "\\bin\\" . $type . "\\.VCS_" . $type . "_" . $self->{'resname'} . "_Online_oldstatefile.DO-NOT-REMOVE";
    }
    else {
        $oldstatefile = "$vcs_home" . "/bin/" . $type . "/.VCS_" . $type . "_" . $self->{'resname'} . "_Online_oldstatefile.DO-NOT-REMOVE";
    }

    my $oldState = "";
    if(open(LFH, "<$oldstatefile")){
        my @lines = <LFH>;
        if ($#lines + 1 > 0) {
            $oldState = $lines[0];
            chomp($oldState);
        }
        close(LFH);
        $self->{'oldstatus'} = $oldState;
    }

	#Initialize if the Oldstate is empty.
	#i.e don't trap the very first time
	#just save the old state
    if ($oldState eq ""){
        $self->{'oldstatus'} = $self->{'status'};
        if(open(LFH, ">$oldstatefile")){
            print LFH $self->{'status'};
            close(LFH);
        }
        else {
            VCSAG_LOGDBG_MSG (3, "setOldState : Failed to  open $oldstatefile file");
        }
    }
}

sub setOldState {
	my ($self) = @_;
    my ($type) = $self->{'resource_type'};
    if (0 == length $type) {
        # Resource type was not set in the $self object
        ($type) = split('\n',`$ag_i18n_inc::HARES -value $self->{'resname'} Type`);
    }


	#if the current state is changed, set old state with current state.
    if ($self->{'oldstatus'} ne $self->{'status'}){

        my $oldstatefile;
        if ( $^O =~ /^MSWin/i ){
            my $vcshome = $ag_i18n_inc::VCSHOME;
            $vcshome =~ s/\"//g;
            $oldstatefile = "$vcshome" . "\\bin\\" . $type . "\\.VCS_" . $type . "_" . $self->{'resname'} . "_Online_oldstatefile.DO-NOT-REMOVE";
        }
        else {
            $oldstatefile = "$vcs_home" . "/bin/" . $type . "/.VCS_" . $type . "_" . $self->{'resname'} . "_Online_oldstatefile.DO-NOT-REMOVE";
        }

        if(open(LFH, ">$oldstatefile")){
            print LFH $self->{'status'};
            close(LFH);
        }
        else{
            VCSAG_LOGDBG_MSG (3, "setOldState : Failed to  open $oldstatefile file");
        }
    }
}

sub removeOldStateFile {
    my ($self) = @_;
    my ($type) = $self->{'resource_type'};
    if (0 == length $type) {
        # Resource type was not set in the $self object
        ($type) = split('\n',`$ag_i18n_inc::HARES -value $self->{'resname'} Type`);
    }

    my $oldstatefile;
    if ( $^O =~ /^MSWin/i ){
        my $vcshome = $ag_i18n_inc::VCSHOME;
        $vcshome =~ s/\"//g;
        $oldstatefile = "$vcshome" . "\\bin\\" . $type . "\\.VCS_" . $type . "_" . $self->{'resname'} . "_Online_oldstatefile.DO-NOT-REMOVE";
    }
    else {
        $oldstatefile = "$vcs_home" . "/bin/" . $type . "/.VCS_" . $type . "_" . $self->{'resname'} . "_Online_oldstatefile.DO-NOT-REMOVE";
    }

    if(-e $oldstatefile){
        unlink($oldstatefile);
    }
}

sub current_role {
    my($self) = @_;
    VCSAG_LOG_MSG("N", " current role : role: [$self->{'role'}] status :[$self->{'status'}]", 6835, "$self->{'role'}", "$self->{'status'}");

    if ( $self->{'role'} eq "P-VOL" ){
        return "P-VOL";
    }
    elsif ( $self->{'role'} eq "SMPL" ){
        return "SMPL";
    }
    elsif ( $self->{'role'} eq "S-VOL" && $self->{'status'} eq "SSWS") {
        return "S-VOL:SSWS";
    }
    elsif ( $self->{'role'} eq "S-VOL" && $self->{'status'} eq "SSUS" && $self->{'isSSUSWritable'} == 1 ){
        return "S-VOL:SSUS:W";
    }
    return "S-VOL";
}
